
class Operation(object):
    """
    Operator precedence determines the grouping of terms
    in an expression and decides how an expression is evaluated.

    Within an expression, higher precedence operators will be evaluated first.
    """
    symbol: str = 'undefined'
    precedence: int = 0


class BooleanOperation(Operation):
    """
    A base class for operations evaluating to a boolean expression
    """
    pass


class LogicalOperation(BooleanOperation):
    pass


class Or(LogicalOperation):
    """
    TRUE if either Boolean expression is TRUE
    """
    symbol = 'OR'
    precedence = 11


class In(LogicalOperation):
    """
    TRUE if the operand is equal to one of a list of expressions
    """
    symbol = 'IN'
    precedence = 12


class NotIn(LogicalOperation):
    symbol = 'NOT IN'
    precedence = 12


class Between(LogicalOperation):
    """
    TRUE if the operand is within a range
    """
    symbol = 'BETWEEN'
    precedence = 13


class NotBetween(LogicalOperation):
    symbol = 'NOT BETWEEN'
    precedence = 13


class Like(LogicalOperation):
    """
    TRUE if the operand matches a pattern
    """
    symbol = 'LIKE'
    precedence = 14


class NotLike(LogicalOperation):
    symbol = 'NOT LIKE'
    precedence = 14


class Ilike(LogicalOperation):
    """
    TRUE if the operand matches a pattern (case-insensitive)
    """
    symbol = 'ILIKE'
    precedence = 14


class And(LogicalOperation):
    """
    TRUE if both Boolean expressions are TRUE
    """
    symbol = 'AND'
    precedence = 15


class EqualityOperation(BooleanOperation):
    pass


class Eq(EqualityOperation):
    """
    Equal to
    """
    symbol = '='
    precedence = 21


class Ne(EqualityOperation):
    """
    Not equal to
    """
    symbol = '!='
    precedence = 21


class RelationalOperation(BooleanOperation):
    pass


class Lt(RelationalOperation):
    """
    Less than
    """
    symbol = '<'
    precedence = 25


class Le(RelationalOperation):
    """
    Less than or equal to
    """
    symbol = '<='
    precedence = 25


class Ge(RelationalOperation):
    """
    Greater than or equal to
    """
    symbol = '>='
    precedence = 25


class Gt(RelationalOperation):
    """
    Greater than
    """
    symbol = '>'
    precedence = 25


class ArithmeticOperation(Operation):
    """
    A base class for operations evaluating to an arithmetic expression
    """
    pass


class AdditiveOperation(ArithmeticOperation):
    pass


class Add(AdditiveOperation):
    """
    Add
    """
    symbol = '+'
    precedence = 31


class Sub(AdditiveOperation):
    """
    Subtract
    """
    symbol = '-'
    precedence = 31


class MultiplicativeOperation(ArithmeticOperation):
    pass


class Mul(MultiplicativeOperation):
    """
    Multiply
    """
    symbol = '*'
    precedence = 35


class Div(MultiplicativeOperation):
    """
    Divide
    """
    symbol = '/'
    precedence = 35


class Mod(MultiplicativeOperation):
    """
    Modulo
    """
    symbol = '%'
    precedence = 35


class UnaryOperation(Operation):
    pass


class USub(UnaryOperation, ArithmeticOperation):
    """
    Negate
    """
    symbol = '-'
    precedence = 41


class Not(UnaryOperation, LogicalOperation):
    """
    Negate
    """
    symbol = 'NOT '
    precedence = 41



import copy
import datetime
from typing import List, Any, Union, Iterable, Optional
from typing import TYPE_CHECKING

from . import operators as op

if TYPE_CHECKING:
    from . import functions as fn


class BaseExpression(object):

    def __init__(
            self,
            body: Any,
            parent: Optional[Any] = None,
            alias_name: Optional[str] = None,
            cast_dtype: Optional[str] = None
            ):
        """
        Parameters
        ----------
        body
            The expression body
        alias_name
            A temporary alias name for the expression
        cast_dtype
            The name of the data type to which expression values are converted
        """
        self._body = body
        self._parent = parent
        self._alias = alias_name
        self._cast = cast_dtype

    @property
    def _body_str(self) -> str:
        """
        Overwritten by child classes

        Returns
        -------
        str
            String representation of the inner body
        """
        return str(self._body)

    @property
    def precedence(self) -> int:
        """
        A large number to ...
        """
        return 100

    @property
    def evaluated_name(self) -> Union[str, None]:
        return self._alias

    @property
    def evaluated_dtype(self) -> Union[str, None]:
        return self._cast

    @staticmethod
    def wrap_parenthesis(s: str) -> str:
        """
        Put parenthesis around the string or leave it unchanged if
        it already has the parenthesis

        Parameters
        ----------
        s
            The input string

        Returns
        -------
        str
            A string enclosed in parenthesis or
        """
        if not s.strip():  # empty string
            return s
        if s.startswith('(') and s.endswith(')'):  # already has parenthesis
            return s
        else:
            return f'({s})'

    def alias(self, alias_name: str) -> 'BaseExpression':
        """
        Old alias is overwritten

        Parameters
        ----------
        alias_name
            A temporary alias name for the column

        Returns
        -------
        Column
            A new instance of BaseExpression with the alias set
        """
        created = copy.copy(self)
        created._alias = alias_name
        return created

    def cast(self, dtype: str) -> 'BaseExpression':
        """
        Parameters
        ----------
        dtype
            A name of the data type

        Returns
        -------
        Column
            A new instance of BaseExpression with the cast data type set
        """
        created = copy.copy(self)
        created._cast = dtype
        return created

    def astype(self, dtype: str) -> 'BaseExpression':
        """
        Same as `cast`
        """
        return self.cast(dtype)

    def sql(self) -> str:
        """
        SQL representation of the expression
        """
        return str(self)

    def _format_str(self, s: str) -> str:
        """
        Apply alias and/or cast data type.
        """
        if self._cast:
            s = f'CAST({s} AS {self._cast})'
        if self._alias:
            s = f'{s} AS "{self._alias}"'
        return s

    def __copy__(self) -> 'BaseExpression':
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        return result

    def __str__(self) -> str:
        return self._format_str(self._body_str)

    def __hash__(self) -> int:
        return hash(str(self))

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self._body})'

    def __and__(self, other: Any) -> 'BooleanExpression':
        return BooleanExpression(op.And(), self, Literal.wrap_constant(other))

    def __or__(self, other: Any) -> 'BooleanExpression':
        return BooleanExpression(op.Or(), self, Literal.wrap_constant(other))

    def __rand__(self, other: Any) -> 'BooleanExpression':
        return BooleanExpression(op.And(), Literal.wrap_constant(other), self)

    def __ror__(self, other: Any) -> 'BooleanExpression':
        return BooleanExpression(op.Or(), Literal.wrap_constant(other), self)

    def __lt__(self, other: Any) -> 'BooleanExpression':
        return BooleanExpression(op.Lt(), self, Literal.wrap_constant(other))

    def __le__(self, other: Any) -> 'BooleanExpression':
        return BooleanExpression(op.Le(), self, Literal.wrap_constant(other))

    def __eq__(self, other: Any) -> 'BooleanExpression':  # type: ignore[override]
        return BooleanExpression(op.Eq(), self, Literal.wrap_constant(other))

    def __ne__(self, other: Any) -> 'BooleanExpression':  # type: ignore[override]
        return BooleanExpression(op.Ne(), self, Literal.wrap_constant(other))

    def __gt__(self, other: Any) -> 'BooleanExpression':
        return BooleanExpression(op.Gt(), self, Literal.wrap_constant(other))

    def __ge__(self, other: Any) -> 'BooleanExpression':
        return BooleanExpression(op.Ge(), self, Literal.wrap_constant(other))

    def __add__(self, other: Any) -> 'ArithmeticExpression':
        return ArithmeticExpression(op.Add(), self, Literal.wrap_constant(other))

    def __sub__(self, other: Any) -> 'ArithmeticExpression':
        return ArithmeticExpression(op.Sub(), self, Literal.wrap_constant(other))

    def __mul__(self, other: Any) -> 'ArithmeticExpression':
        return ArithmeticExpression(op.Mul(), self, Literal.wrap_constant(other))

    def __truediv__(self, other: Any) -> 'ArithmeticExpression':
        return ArithmeticExpression(op.Div(), self, Literal.wrap_constant(other))

    def __mod__(self, other: Any) -> 'ArithmeticExpression':
        return ArithmeticExpression(op.Mod(), self, Literal.wrap_constant(other))

    def __radd__(self, other: Any) -> 'ArithmeticExpression':
        return ArithmeticExpression(op.Add(), Literal.wrap_constant(other), self)

    def __rsub__(self, other: Any) -> 'ArithmeticExpression':
        return ArithmeticExpression(op.Sub(), Literal.wrap_constant(other), self)

    def __rmul__(self, other: Any) -> 'ArithmeticExpression':
        return ArithmeticExpression(op.Mul(), Literal.wrap_constant(other), self)

    def __rdiv__(self, other: Any) -> 'ArithmeticExpression':
        return ArithmeticExpression(op.Div(), Literal.wrap_constant(other), self)

    def __rmod__(self, other: Any) -> 'ArithmeticExpression':
        return ArithmeticExpression(op.Mod(), Literal.wrap_constant(other), self)

    def __neg__(self) -> 'UnaryExpression':
        return UnaryExpression(op.USub(), self)

    def __pos__(self) -> 'BaseExpression':
        return self

    def __abs__(self) -> 'UnaryExpression':
        raise NotImplementedError

    def __invert__(self) -> 'UnaryExpression':
        return UnaryExpression(op.Not(), self)

    def lt(self, other: Any) -> 'BooleanExpression':
        return self < other

    def le(self, other: Any) -> 'BooleanExpression':
        return self <= other

    def eq(self, other: Any) -> 'BooleanExpression':
        return self == other

    def ne(self, other: Any) -> 'BooleanExpression':
        return self != other

    def gt(self, other: Any) -> 'BooleanExpression':
        return self > other

    def ge(self, other: Any) -> 'BooleanExpression':
        return self >= other

    def add(self, other: Any) -> 'ArithmeticExpression':
        return self + other

    def sub(self, other: Any) -> 'ArithmeticExpression':
        return self - other

    def mul(self, other: Any) -> 'ArithmeticExpression':
        return self * other

    def div(self, other: Any) -> 'ArithmeticExpression':
        return self / other

    def mod(self, other: Any) -> 'ArithmeticExpression':
        return ArithmeticExpression(op.Mod(), self, Literal.wrap_constant(other))

    def isin(self, other: Iterable[Any]) -> 'BooleanExpression':
        return BooleanExpression(op.In(), self, Literal.wrap_constant(tuple(other)))

    def like(self, other: str) -> 'BooleanExpression':
        return BooleanExpression(op.Like(), self, Literal.wrap_constant(other))

    def ilike(self, other: str) -> 'BooleanExpression':
        return BooleanExpression(op.Ilike(), self, Literal.wrap_constant(other))

    def between(self, beg: Any, end: Any) -> 'BooleanExpression':
        other = Literal.wrap_constant(beg) & Literal.wrap_constant(end)
        return BooleanExpression(op.Between(), self, other)

    def apply(self, func: 'fn.Function', *args) -> 'FuncExpression':
        # if isinstance(func, str):
        #     func = Function(func)
        return func(self, *args)


class Expression(BaseExpression):
    def __init__(
            self,
            operation: op.Operation,
            operands: List[BaseExpression],
            parent: Optional[Any] = None,
            alias_name: str = None,
            cast_dtype: str = None
            ):
        body = self.op, self.operands = operation, operands
        super().__init__(body, parent, alias_name, cast_dtype)

    @property
    def precedence(self) -> int:
        """
        """
        return self.op.precedence


class BinaryExpression(Expression):
    def __init__(
            self,
            operation: op.Operation,
            left: BaseExpression,
            right: BaseExpression,
            parent: Optional[Any] = None,
            alias_name: str = None,
            cast_dtype: str = None
            ):
        self.op, self.left, self.right = operation, left, right
        super().__init__(operation, [left, right], parent, alias_name, cast_dtype)

    def _infer_parent(self) -> Any:
        if self.left._parent is None or self.right._parent is None:
            return self.left._parent or self.right._parent
        elif self.left._parent is self.right._parent:
            return self.left._parent
        else:
            msg = (
                "Cannot use operands from distinct datasets: " +
                f"{self.left._parent} and {self.right._parent}"
            )
            raise ValueError(msg)

    @property
    def _body_str(self) -> str:
        left_str = str(self.left)
        right_str = str(self.right)
        if self.precedence > self.left.precedence:
            left_str = self.wrap_parenthesis(left_str)
        if self.precedence > self.right.precedence:
            right_str = self.wrap_parenthesis(right_str)
        if isinstance(self.right, UnaryExpression):
            right_str = self.wrap_parenthesis(right_str)
        if isinstance(self.left, BooleanExpression):
            left_str = self.wrap_parenthesis(left_str)
        if isinstance(self.right, BooleanExpression):
            right_str = self.wrap_parenthesis(right_str)
        return f"{left_str} {self.op.symbol} {right_str}"


class BooleanExpression(BinaryExpression):
    @property
    def _body_str(self) -> str:
        if isinstance(self.op, (op.Between, op.NotBetween)):
            left_str = str(self.left)
            right_str = str(self.right)
            if self.precedence > self.left.precedence:
                left_str = self.wrap_parenthesis(left_str)
            return f"{left_str} {self.op.symbol} {right_str}"
        else:
            return super()._body_str


class ArithmeticExpression(BinaryExpression):
    pass


class UnaryExpression(Expression):

    def __init__(
            self,
            operation: op.Operation,
            operand: BaseExpression,
            parent: Optional[Any] = None,
            alias_name: str = None,
            cast_dtype: str = None
            ):
        self.op, self.operand = operation, operand
        super().__init__(self.op, [self.operand], parent, alias_name, cast_dtype)

    @property
    def _body_str(self) -> str:
        operand_str = str(self.operand)
        if self.precedence > self.operand.precedence:
            operand_str = self.wrap_parenthesis(operand_str)
        return f"{self.op.symbol}{operand_str}"


class FuncExpression(Expression):

    def __init__(
            self,
            function: 'fn.Function',
            operand: List[BaseExpression],
            parent: Optional[Any] = None,
            alias_name: str = None,
            cast_dtype: str = None
            ):
        self.fn, self.operand = function, operand
        super().__init__(function, operand, parent, alias_name, cast_dtype)

    @property
    def _body_str(self) -> str:
        operand_str = ", ".join(str(x) for x in self.operand)
        return f"{self.fn.symbol}({operand_str})"


class Literal(BaseExpression):
    """
    Represents a constant value
    """
    @classmethod
    def wrap_constant(cls, value: Any) -> BaseExpression:
        """
        If not an Expression, make it a literal
        """
        return value if isinstance(value, BaseExpression) else Literal(value)

    @property
    def value(self):
        return self._body

    def _format_literal(self, value: Any) -> str:
        if value is None:
            return "NULL"
        elif isinstance(value, bool):
            return str(value).upper()
        elif isinstance(value, (datetime.date, datetime.datetime)):
            return f"'{value.isoformat()}'"
        elif isinstance(value, (list, tuple, set)):
            return '(' + ", ".join(map(self._format_literal, value)) + ')'
        elif isinstance(value, (int, float)):
            return str(value)
        else:
            return f"'{value}'"

    @property
    def _body_str(self) -> str:
        return self._format_literal(self.value)



import os
from typing import Callable, Optional, Any

from clickhouse_driver import Client
from clickhouse_driver.errors import Error


class ExecutionError(Exception):
    pass


class Session(object):

    def __init__(
            self,
            host: Optional[str] = None,
            user: Optional[str] = None,
            password: Optional[str] = None,
            port: Optional[int] = None,
            secure: bool = False,
            **kwargs: Any
            ):
        default_port = 9000
        self._host = host or os.getenv('AS_DATABASE_HOST')
        self._port = port or os.getenv('AS_DATABASE_PORT') or default_port
        self._user = user or os.getenv('AS_DATABASE_USER')
        self._password = password or os.getenv('AS_DATABASE_PASSWORD')
        self._secure = secure

        self.client = Client(
            self._host,
            user=self._user,
            password=self._password,
            port=self._port,
            secure=self._secure,
            **kwargs
        )

    def ping(self):
        return self.execute('SELECT 1')

    def close(self):
        self.client.disconnect()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def _wrap_exception(self, func: Callable, query: str, **kwargs: Any):
        try:
            return func(query, **kwargs)
        except Error as e:
            message = e.message.split('Stack trace:')[0].split(': ', 2)[-1]
            raise ExecutionError(message)

    def execute(self, query: str, **kwargs: Any):
        return self._wrap_exception(self.client.execute, query, **kwargs)

    def query_dataframe(self, query: str, **kwargs: Any):
        return self._wrap_exception(self.client.query_dataframe, query, **kwargs)

    def __hash__(self) -> int:
        return hash(f"{self._host}{self._user}{self._password}")



from .operators import Operation
from .expressions import BaseExpression, FuncExpression, Literal


class Function(Operation):
    """
    """

    precedence = 90

    def __init__(self, name: str):
        self.name = name
        self.symbol = name

    def __call__(self, x: 'BaseExpression', *args, **kwargs) -> 'FuncExpression':
        # TODO: cannot apply a function to column by its string name
        if not isinstance(x, BaseExpression):
            x = Literal.wrap_constant(x)
        return FuncExpression(self, [x] + list(args))


class AggregateFunction(Function):
    pass


distinct = Function('distinct')
max2 = Function('max2')
min2 = Function('min2')

toInt = Function('toInt')
toUInt = Function('toUInt')
toFloat = Function('toFloat')
toDecimal = Function('toDecimal')

toDate = Function('toDate')
toDateTime = Function('toDateTime')
toTime = Function('toTime')
toString = Function('toString')

toIntervalSecond = Function('toIntervalSecond')
toIntervalMinute = Function('toIntervalMinute')
toIntervalHour = Function('toIntervalHour')
toIntervalDay = Function('toIntervalDay')
toIntervalWeek = Function('toIntervalWeek')
toIntervalMonth = Function('toIntervalMonth')
toIntervalQuarter = Function('toIntervalQuarter')
toIntervalYear = Function('toIntervalYear')

toUnixTimestamp64Milli = Function('toUnixTimestamp64Milli')
toUnixTimestamp64Micro = Function('toUnixTimestamp64Micro')
toUnixTimestamp64Nano = Function('toUnixTimestamp64Nano')
fromUnixTimestamp64Milli = Function('fromUnixTimestamp64Milli')
fromUnixTimestamp64Micro = Function('fromUnixTimestamp64Micro')
fromUnixTimestamp64Nano = Function('fromUnixTimestamp64Nano')

cast = Function('cast')
abs = Function('abs')

sum = AggregateFunction('sum')
avg = AggregateFunction('avg')
mean = AggregateFunction('mean')
count = AggregateFunction('count')
max = AggregateFunction('max')
min = AggregateFunction('min')
argMax = AggregateFunction('argMax')
argMin = AggregateFunction('argMin')
any = AggregateFunction('any')
avgWeighted = AggregateFunction('avgWeighted')
groupArray = AggregateFunction('groupArray')
uniq = AggregateFunction('uniq')



#!/usr/bin/env python
import os
import sys

from setuptools import setup
from typing import Dict

# 'setup.py publish' shortcut.
if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist bdist_wheel')
    os.system('twine upload dist/*')
    sys.exit()


about: Dict[str, str] = {}
with open("algoseek_connector/__version__.py") as fp:
    exec(fp.read(), about)
with open("README.md") as fp:
    readme = fp.read()


requires = [
    'clickhouse-driver',
    'pandas'
]

setup(
    name=about['__title__'],
    version=about['__version__'],
    description=about['__description__'],
    long_description=readme,
    long_description_content_type='text/markdown',
    author=about['__author__'],
    author_email=about['__author_email__'],
    packages=['algoseek_connector'],
    package_data={'': ['LICENSE', 'NOTICE']},
    package_dir={'algoseek_connector': 'algoseek_connector'},
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: SQL',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Database',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    include_package_data=True,
    python_requires=">=3.6, <4",
    install_requires=requires,
    zip_safe=False
)



from .__version__ import __title__, __description__, __version__
from .__version__ import __author__, __author_email__, __copyright__

from .sessions import Session, ExecutionError
from .resources import DataResource, Datagroup, Dataset

__all__ = [
    '__title__', '__description__', '__version__',
    '__author__', '__author_email__', '__copyright__',
    'Session', 'ExecutionError', 'DataResource', 'Datagroup', 'Dataset'
]



import copy
import functools
from numbers import Number
from typing import List, Tuple, Iterable, Iterator, Optional, Union, Sequence, Any

from .sessions import Session
from .expressions import BaseExpression, Expression, Literal


@functools.lru_cache()
def list_datagroups(session: Session) -> Tuple[str]:
    return session.execute(
        "SHOW DATABASES", columnar=True
    )[0]


class DatagroupsMeta(object):

    def __init__(self, session: Optional[Session] = None):
        self.session = session or Session()
        self.names = list_datagroups(self.session)


@functools.lru_cache()
def list_datasets(datagroup_name: str, session: Session) -> Tuple[str]:

    # TODO:
    # datagroup_name in list_datagroups()
    return session.execute(
        f"SHOW TABLES FROM {datagroup_name}", columnar=True
    )[0]


class DatasetsMeta(object):

    def __init__(self, datagroup_name: str, session: Optional[Session] = None):
        self.session = session or Session()
        self.names = list_datasets(datagroup_name, self.session)


@functools.lru_cache()
def list_columns(
        datagroup_name: str,
        dataset_name: str,
        session: Session
        ) -> Tuple[Tuple[str], Tuple[str], Tuple[str]]:

    names, types, _, _, descriptions, _, _ = session.execute(
        f"DESCRIBE TABLE {datagroup_name}.{dataset_name}",
        columnar=True
    )
    return names, tuple(parse_column_type(x) for x in types), descriptions


def parse_column_type(type_name: str) -> str:
    # TODO
    if type_name.startswith('Enum'):
        return 'String'
    elif type_name.startswith('LowCardinality'):
        return type_name.split('(', 1)[1].rsplit(')', 1)[0]
    else:
        return type_name


def is_iterable_type(value: Any) -> bool:
    return isinstance(value, Sequence) and not isinstance(value, str)


class ColumnsMeta(object):

    def __init__(
            self,
            datagroup_name: str,
            dataset_name: str,
            session: Optional[Session] = None
            ):
        self.session = session or Session()
        self.names, self.types, self.descriptions = list_columns(
            datagroup_name, dataset_name, self.session
        )

    def __iter__(self) -> Iterator[Tuple[str, str, str]]:
        for triplet in zip(self.names, self.types, self.descriptions):
            yield triplet


class Datagroup(object):

    def __init__(
            self, name: str,
            session: Optional[Session] = None,
            meta: Optional[DatagroupsMeta] = None
            ):
        self.name = name
        self.session = session or Session()
        self.meta = meta or DatagroupsMeta(self.session)
        if name not in self.meta.names:
            raise ValueError(f"Cannot identify a datagroup with a name '{name}'")

    @property
    def datasets(self) -> 'DatasetSelector':
        return DatasetSelector(self, self.session)

    @property
    def _datasets_meta(self) -> 'DatasetsMeta':
        return DatasetsMeta(self.name, self.session)

    def __str__(self) -> str:
        return f"Datagroup({self.name})"


class DatagroupSelector(object):

    def __init__(self, session: Optional[Session] = None):

        self.session = session or Session()
        self.meta = DatagroupsMeta(self.session)
        self._datagroups = {
            name: Datagroup(
                name, session=self.session, meta=self.meta
            )
            for name in self.meta.names
        }
        for name, datagroup in self._datagroups.items():
            setattr(self, name, datagroup)

    def all(self) -> Iterable[Datagroup]:
        for datagroup in self._datagroups.values():
            yield datagroup


DatagroupIdentifier = Union[str, Datagroup]
ExprIdentifier = Union[str, BaseExpression]


class Dataset(object):

    def __init__(
            self,
            datagroup: DatagroupIdentifier,
            name: Optional[str] = None,
            session: Optional[Session] = None
            ):

        if isinstance(datagroup, str):
            if not name and '.' in datagroup:
                datagroup, name = datagroup.split('.', 1)
            self.datagroup = Datagroup(datagroup, session=session)
        elif isinstance(datagroup, Datagroup):
            self.datagroup = datagroup
        else:
            raise ValueError(f"Cannot cast {datagroup} into a datagroup")

        if name not in self.datagroup._datasets_meta.names:
            raise ValueError(f"Cannot identify a dataset with a name '{name}'")

        self.session = self.datagroup.session
        self.name = name
        self.meta = ColumnsMeta(self.datagroup.name, name, self.session)
        self._columns = {
            name: Column(name, dtype, desr, parent=self)
            for name, dtype, desr in self.meta
        }
        for name, column in self._columns.items():
            setattr(self, name, column)
        self.expression = DatasetExpression(self)

    def __str__(self) -> str:
        return f"Dataset({self.datagroup.name}.{self.name})"

    def select(self, *args: Any) -> 'Dataset':
        result = self.copy()
        result.expression.select(*args)
        return result

    def filter(self, expr: Expression) -> 'Dataset':
        result = self.copy()
        result.expression.filter(expr)
        return result

    def groupby(self, *args: Any) -> 'Dataset':
        result = self.copy()
        result.expression.groupby(*args)
        return result

    def sort(self, *args: Any) -> 'Dataset':
        result = self.copy()
        result.expression.sort(*args)
        return result

    def head(self, n: int = 10) -> 'Dataset':
        result = self.copy()
        result.expression.head(n)
        return result

    def fetch(self):
        df = self.session.query_dataframe(str(self.expression))
        return df

    def sql(self) -> str:
        return str(self.expression)

    def copy(self) -> 'Dataset':
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        result.expression = copy.copy(self.expression)
        result.expression._select = self.expression._select.copy()
        result.expression._groupby = self.expression._groupby.copy()
        result.expression._sort = self.expression._sort.copy()
        return result


class DatasetSelector(object):

    def __init__(
            self,
            datagroup: DatagroupIdentifier,
            session: Optional[Session] = None
            ):

        if isinstance(datagroup, str):
            self.datagroup = Datagroup(datagroup, session=session)
        elif isinstance(datagroup, Datagroup):
            self.datagroup = datagroup
        else:
            raise ValueError(f"Cannot cast {datagroup} into a datagroup")

        self.session = session or self.datagroup.session
        self.meta = self.datagroup._datasets_meta
        self._datasets = {
            name: Dataset(self.datagroup, name)
            for name in self.meta.names
        }
        for name, dataset in self._datasets.items():
            setattr(self, name, dataset)

    def all(self) -> Iterable['Dataset']:
        for dataset in self._datasets.values():
            yield dataset


class DatasetExpression(object):

    def __init__(self, parent: Dataset):
        self.parent = parent
        self._namespace = set(self.parent._columns.keys())
        self._select: List[BaseExpression] = []
        self._filter: Optional[Expression] = None
        self._groupby: List[ExprIdentifier] = []
        self._sort: List[ExprIdentifier] = []
        self._limit: Optional[int] = None

    @property
    def _table_name(self) -> str:
        return f"{self.parent.datagroup.name}.{self.parent.name}"

    def _parse_identifier(self, item) -> BaseExpression:
        if isinstance(item, BaseExpression):
            value = item
        elif isinstance(item, str) and item in self.parent._columns:
            value = self.parent._columns[item]
        elif isinstance(item, (Number, bool, str)):
            value = Literal.wrap_constant(item)
        else:
            raise ValueError(f"Cannot parse the value: {item}")
        return value

    def select(self, *args: Any):
        # columns will be added next or raise an error
        columns = []
        if not args:
            raise ValueError("Cannot select with an empty sequence")
        if len(args) == 1 and (isinstance(args[0], dict) or is_iterable_type(args[0])):
            targets = args[0]
        else:
            targets = args

        if isinstance(targets, Sequence):
            for item in targets:
                value = self._parse_identifier(item)
                columns.append(value)
        elif isinstance(targets, dict):
            for key, item in targets.items():
                value = self._parse_identifier(item)
                columns.append(value.alias(key))
        else:
            raise ValueError(f"Invalid type for select expression: {type(targets)}")

        self._select.clear()
        for item in columns:
            if isinstance(item, BaseExpression):
                self._select.append(item)
                if item.evaluated_name is not None:
                    self._namespace.add(item.evaluated_name)
                # TODO:
                # if item._parent is self:
                #     self._select.append(item)
                # else:
                #     raise ValueError("")
            elif isinstance(item, str):
                self._select.append(self.parent._columns[item])
            else:
                self._select.append(Literal.wrap_constant(item))

    def filter(self, expr: 'Expression'):
        # TODO: validate expr
        self._filter = expr if self._filter is None else self._filter & expr

    def head(self, n: int):
        if not isinstance(n, int):
            raise ValueError(f'An integer value is expected: got {type(n)} instead')
        if n <= 0:
            raise ValueError(f'A positive integer is expected: got {n} instead')
        self._limit = n

    def groupby(self, *args: Any):
        targets = args[0] if len(args) == 1 and is_iterable_type(args[0]) else args
        # TODO: validate columns
        for item in targets:
            if isinstance(item, str):
                if item not in self._namespace:
                    raise ValueError(f"{item} not in namespace")
                else:
                    self._groupby.append(item)
            else:
                self._groupby.append(item)

    def sort(self, *args: Any):
        targets = args[0] if len(args) == 1 and is_iterable_type(args[0]) else args
        # TODO: validate columns
        for item in targets:
            if isinstance(item, str):
                if item not in self._namespace:
                    raise ValueError(f"{item} not in namespace")
                else:
                    self._sort.append(item)
            else:
                self._sort.append(item)

    def __str__(self) -> str:
        select = '*'
        if self._select:
            select = ', '.join(col.sql() for col in self._select)
        sql_expr = ''
        sql_expr += f"SELECT {select}"
        sql_expr += f'\nFROM {self._table_name}'
        if self._filter:
            sql_expr += f'\nWHERE {self._filter.sql()}'
        if self._groupby:
            groupby_expr = ', '.join(map(str, self._groupby))
            sql_expr += f'\nGROUP BY {groupby_expr}'
        if self._sort:
            sort_expr = ', '.join(map(str, self._sort))
            sql_expr += f'\n ORDER BY {sort_expr}'
        if self._limit:
            sql_expr += f'\nLIMIT {self._limit}'
        return sql_expr


class Column(BaseExpression):
    """
    Represents an individual database column or column alias
    """

    def __init__(
            self,
            name: str,
            dtype: str,
            descr: Optional[str] = None,
            parent: Optional[Dataset] = None,
            alias_name: Optional[str] = None,
            cast_dtype: Optional[str] = None
            ):
        """

        Parameters
        ----------
        name
            The column name
        dtype
            The data type of values in the column
        descr
            A text describing the column contents
        alias_name
            A temporary alias name for the column
        cast_dtype
            The name of the data type to which column values are converted
        """
        body = self.name, self.dtype, self.description = name, dtype, descr
        super().__init__(body, parent, alias_name, cast_dtype)
        if self._parent and name not in self._parent.meta.names:
            raise ValueError(f"A dataset '{self._parent.name}' doesn't have a column '{name}'")

    @property
    def _body_str(self) -> str:
        return f'"{self.name}"'

    @property
    def evaluated_name(self) -> str:
        return self._alias or self.name

    @property
    def evaluated_dtype(self) -> str:
        return self._cast or self.dtype


class DataResource(object):

    def __init__(self, session: Optional[Session] = None):
        self.session = session or Session()

    @property
    def datagroups(self) -> DatagroupSelector:
        return DatagroupSelector(session=self.session)

    def datagroup(self, name: str) -> Datagroup:
        return Datagroup(name, session=self.session)



__title__ = 'algoseek-connector'
__description__ = 'A wrapper library for ORM-like SQL builder and executor'
__version__ = '1.0.2'
__author__ = 'Taras Kuzyo'
__author_email__ = 'taras@algoseek.com'
__copyright__ = 'Copyright 2022 algoseek, LLC'
__license__ = "Apache 2.0"
